/* 
    TVscreen Widget for Bt848 frame grabber driver

    Copyright (C) 1996,97 Marcus Metzler (mocm@thp.uni-koeln.de)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
    Much hacked by Alan Cox 10/1997 in order to work with the development
    version of the "linux video capture" interface. No real algorithm
    changes here, its basically rewrapping the ioctl calls. This will
    still also really only handle the bttv card right now.
*/

/*
 * TVscreen.c 
 */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/IntrinsicP.h>
#include <X11/ObjectP.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>


#define __KERNEL__
#include <linux/videodev.h>
#undef __KERNEL__
#include "../bttv.h"

#include <stdio.h>

#include "TVscreenP.h"
#include "visual.h"

//#define DEBUG  
#define DEFAULT_WIDTH    320
#define DEFAULT_HEIGHT   240  

#define PAL_WIDTH 768
#define PAL_HEIGHT 576

#define NTSC_WIDTH 640
#define NTSC_HEIGHT 480

#define DEFAULT_DELAY 100

#define offset(field) XtOffsetOf(TVscreenRec, field)


static XtResource resources[] = {
  {
    XtNDevname, 
    XtCDevname, 
    XtRString, 
    sizeof(String),
    offset(TVscreen.devname), 
    XtRImmediate, 
    NULL
  },
  {
    XtNCAP, 
    XtCCAP, 
    XtRBoolean, 
    sizeof(Boolean),
    offset(TVscreen.CAP_ON), 
    XtRImmediate, 
    (XtPointer) False
  },
  {
    XtNNORM, 
    XtCNORM, 
    XtRInt, 
    sizeof(int),
    offset(TVscreen.NORM), 
    XtRImmediate, 
    (XtPointer) DNORM
  },
  {
    XtNFREQ, 
    XtCFREQ, 
    XtRInt, 
    sizeof(int),
    offset(TVscreen.FREQ), 
    XtRImmediate, 
    (XtPointer) 0
  },
  {
    XtNINPUT, 
    XtCINPUT, 
    XtRInt, 
    sizeof(int),
    offset(TVscreen.INPUT), 
    XtRImmediate, 
    (XtPointer) DIN
  },
  {
    XtNDELAY, 
    XtCDELAY, 
    XtRInt, 
    sizeof(int),
    offset(TVscreen.DELAY), 
    XtRImmediate, 
    (XtPointer) DEFAULT_DELAY
  },
  {
    XtNBright, 
    XtCBright, 
    XtRInt, 
    sizeof(int),
    offset(TVscreen.Bright), 
    XtRImmediate, 
    (XtPointer) DBRI
  },
  {
    XtNCol, 
    XtCCol, 
    XtRInt, 
    sizeof(int),
    offset(TVscreen.Color), 
    XtRImmediate, 
    (XtPointer) DCOL
  },
  {
    XtNHue, 
    XtCHue, 
    XtRInt, 
    sizeof(int),
    offset(TVscreen.Hue), 
    XtRImmediate, 
    (XtPointer) DHUE
  },
  {
    XtNContrast, 
    XtCContrast, 
    XtRInt, 
    sizeof(int),
    offset(TVscreen.Contrast), 
    XtRImmediate, 
    (XtPointer) DCONTRAST
  },
  {
    XtNREFRESH, 
    XtCREFRESH, 
    XtRBoolean, 
    sizeof(Boolean),
    offset(TVscreen.REFRESH), 
    XtRImmediate, 
    (XtPointer) True
  },
  {
    XtNbpp, 
    XtCbpp, 
    XtRInt, 
    sizeof(int),
    offset(TVscreen.bpp), 
    XtRImmediate, 
    (XtPointer) 0
  },
  {
    XtNvis, 
    XtCvis, 
    XtRString, 
    sizeof(String),
    offset(TVscreen.visname), 
    XtRImmediate, 
    NULL
  },
  {
    XtNsnap, 
    XtCsnap, 
    XtRBoolean, 
    sizeof(Boolean),
    offset(TVscreen.snap), 
    XtRImmediate, 
    (XtPointer) False
  },
  {
    XtNrealw, 
    XtCrealw, 
    XtRInt, 
    sizeof(int),
    offset(TVscreen.realw), 
    XtRImmediate, 
    (XtPointer) 0
  },
};

/* Declaration of methods */

static void Initialize();
static void Realize();
static Boolean SetValues();
static void destroy();
static XtGeometryResult QueryGeometry();
static int rx,ry,rh,rw;

/* the following are actions of TVscreen */
static void pConfig();
static void addclip();
static void maketab();
static int getclip();
static void bt848_dma();
static void SetGeo();
static void cap();
static void setmaxsize();
static void ref();
static void refresh();
XErrorHandler xerror();

static Window lastrwin;
static XtIntervalId timer;
static Boolean Move=False;

TVscreenClassRec tvscreenClassRec = {
    {
    /* core_class fields */
    /* superclass        */ (WidgetClass) &coreClassRec,
    /* class_name        */ "TVscreen",
    /* widget_size       */ sizeof(TVscreenRec),
    /* class_initialize      */ NULL,
    /* class_part_initialize     */ NULL,
    /* class_inited          */ FALSE,
    /* initialize        */ Initialize,
    /* initialize_hook       */ NULL,
    /* realize           */ Realize,
    /* actions           */ NULL,
    /* num_actions       */ 0,
    /* resources         */ resources,
    /* num_resources         */ XtNumber(resources),
    /* xrm_class         */ NULLQUARK,
    /* compress_motion       */ TRUE,
    /* compress_exposure     */ XtExposeCompressMultiple,
    /* compress_enterleave   */ TRUE,
    /* visible_interest      */ FALSE,
    /* destroy           */ destroy,
    /* resize            */ NULL,
    /* expose            */ NULL,
    /* set_values        */ SetValues,
    /* set_values_hook       */ NULL,
    /* set_values_almost     */ XtInheritSetValuesAlmost,
    /* get_values_hook       */ NULL,
    /* accept_focus      */ NULL,
    /* version           */ XtVersion,
    /* callback_private      */ NULL,
    /* tm_table          */ NULL,
    /* query_geometry        */ QueryGeometry,
    /* display_accelerator       */ XtInheritDisplayAccelerator,
    /* extension                 */ NULL
    },
    {
    /* extension          */        0,
    },
};

WidgetClass tvscreenWidgetClass = (WidgetClass) & tvscreenClassRec;

/* Realize */
static void Realize(Widget w,
		    XtValueMask *value_mask,
		    XSetWindowAttributes *attributes)
{
	TVscreenWidget tw=(TVscreenWidget) w;
    
	XtCreateWindow(w, (unsigned int) InputOutput,
		 tw->TVscreen.visual, *value_mask, attributes);

	if (tw->TVscreen.CAP_ON)
	{
		int v=tw->TVscreen.INPUT;
		tw->TVscreen.Change=TRUE;
		if(ioctl(tw->TVscreen.fbttv, VIDIOCSWIN, &tw->TVscreen.vwin))
			perror("VIDIOCSWIN");
		/* Set the input */
		if(ioctl(tw->TVscreen.fbttv, VIDIOCSCHAN, &v))
			perror("VIDIOCSTUNER");
		/* Set the frequency */
		if(ioctl(tw->TVscreen.fbttv, VIDIOCSFREQ, &tw->TVscreen.frequency))
			perror("VIDIOCSFREQ");
	}
} 


XErrorHandler xerror(Display *dpy, XErrorEvent *event)
{
	switch (event->error_code)
	{
		case BadWindow:
#ifdef DEBUG
			printf("Bad Window ingnored\n");
#endif
			return 0 ;
    
		case BadMatch:
#ifdef DEBUG
			printf("BadMatch ingnored\n");
#endif
			return 0 ;

		case BadColor:
#ifdef DEBUG
			printf("BadColor ingnored\n");
#endif
			return 0 ;

		default:
			return 0;
	}
}

void ref(TVscreenWidget w)
{
	int scr,depth;
	XSetWindowAttributes xswa;
	unsigned long mask;
	Window rwin,rootW;
	Display *disp;
  
	if (w->TVscreen.REFRESH)
	{
		disp=XtDisplay((Widget)w);
		scr=DefaultScreen(disp);
		rootW= RootWindow(disp, scr);
		depth=w->core.depth;
    
		xswa.override_redirect = True;
		xswa.backing_store = NotUseful;
		xswa.save_under = False;
		xswa.background_pixel=BlackPixel(disp,scr);
		xswa.backing_store=NotUseful;
		mask = (CWBackPixel | CWSaveUnder | CWBackingStore|
			CWOverrideRedirect );
		if(Move)
		{
			rwin = XCreateWindow(disp,rootW,0, 0, 
				w->TVscreen.vbuf.width,w->TVscreen.vbuf.height, 0,
				CopyFromParent, InputOutput, CopyFromParent,
				mask, &xswa);
			Move=False;
		}
		else
		{
			rwin = XCreateWindow(disp,rootW,
				w->TVscreen.vwin.x-80, w->TVscreen.vwin.y-80, 
				w->TVscreen.vwin.width+80,w->TVscreen.vwin.height+80,
				0,CopyFromParent, InputOutput, CopyFromParent,
				mask, &xswa);
		}
		lastrwin=rwin;
		XMapWindow(disp, rwin);
		XUnmapWindow(disp, rwin);
		XDestroyWindow(disp, rwin);
	}
}

 
static void refresh(w)
TVscreenWidget w;
{
	if (timer)
		XtRemoveTimeOut(timer);

	timer=XtAppAddTimeOut(XtWidgetToApplicationContext((Widget)w),
		w->TVscreen.DELAY,(XtTimerCallbackProc)ref,w);
}



static void addclip(cr,ncr,x,y,x2,y2)
struct video_clip *cr;
int *ncr; 
uint x,y,x2,y2;
{
	cr[*ncr].x=x;
	cr[*ncr].y=y;
	cr[*ncr].width=x2-x;
	cr[*ncr].height=y2-y;
	*ncr=*ncr+1;
} 


static int getclip(w)
TVscreenWidget w;
{
	int ncr=0;
	int x,y,x2,y2,wx,wy;
	uint ww,wh;
	struct video_window vwin;
	struct video_buffer vbuf;
	Display *disp;
	XWindowAttributes wts;
	Window parent,win,parent2, root, rroot, *children, swin;
	uint nchildren, i;
	struct video_clip *cr; 
	Widget widpar;


	if (!XtIsRealized((Widget)w))
		return;

	disp=XtDisplay((Widget)w);
	vwin=w->TVscreen.vwin;
	vbuf=w->TVscreen.vbuf;

	widpar=(Widget) w;
	while(XtParent(widpar))
		widpar=XtParent(widpar);
	if((win=XtWindow(widpar))==None)
		return;

	XQueryTree(disp,win, &rroot, &parent, &children, &nchildren);
	if(nchildren)
		XFree((char *)children);

	cr=w->TVscreen.cliprecs;

	/*
	 *	Clip to desktop - note this is a list of rectangles
	 *	_not_ to draw.
	 */
	 
	wx=vwin.x; wy=vwin.y;
	wh=vwin.height; ww=vwin.width;
	
#if 0
	/*
	 *	Clip areas off the top and or left - FIXME - driver
	 *	must be made to do these.
	 */
	 
	if (wx<0)
		addclip(cr, &ncr, 0, 0, (uint)(-wx), wh-1);
	if (wy<0)
		addclip(cr, &ncr, 0, 0, ww-1, (uint)(-wy));
	/*
	 *	Clip bottom right - again driver must do these
	 */
	if ((wx+ww) > vbuf.width)
		addclip(cr, &ncr, vbuf.width-wx, 0, ww-1, wh-1);
	if ((wy+wh) > vbuf.height)
		addclip(cr, &ncr, 0, vbuf.height-wy, ww-1, wh-1);

#endif
	/*
	 *	Find the root window
	 */
	     
	root=DefaultRootWindow(disp);
	swin=win;
	
	/*
	 *	Walk upwards to find our parent
	 */
	 
	while (parent!=root) 
	{
		swin=parent;
		XQueryTree(disp, swin, &rroot, &parent, &children, &nchildren);
		if (nchildren)
			XFree((char *) children);
	}
	XQueryTree(disp, root, &rroot, &parent2, &children, &nchildren);
	
	/*
	 *	Find our top level window
	 */
	 
	for (i=0; i<nchildren;i++) 
		if (children[i]==swin)
			break;
	i++;
	
	/*
	 *	Now clip 
	 */
	 
	for (; i<nchildren; i++) 
	{
		XGetWindowAttributes(disp, children[i], &wts);
		/*
		 *	Not displayed .. no clip
		 */
		if (!(wts.map_state & IsViewable))
			continue;

		/* Bounding rectangle (x,y)-(x2,y2)
		   relative to our window base */
		
		x=wts.x-vwin.x; y=wts.y-vwin.y;
		x2=x+wts.width+2*wts.border_width-1;
		y2=y+wts.height+2*wts.border_width-1;
		
		/*
		 *	Check if the rectangle is actually out of
		 *	drawing area.
		 */
		 
		if ((x2 < 0) || (x >= (int)ww) ||
			(y2 < 0) || (y >= (int)wh))
			continue;
			
		/*
		 *	Clip the rectangle to the drawing area
		 */

		if (x<0)
			x=0;
		if (y<0)
			y=0;
		if (x2>=(int)ww)
			x2=ww-1;
		if (y2>=(int)wh)
			y2=wh-1;
		
		/*
		 *	Throw it on the clip list.
		 */
		addclip(cr, &ncr, x, y, x2, y2);
	}
	XFree((char *) children);
	
	w->TVscreen.vwin.clips=w->TVscreen.cliprecs;
	w->TVscreen.vwin.clipcount=ncr;
	
	/*
	 *	Set the clipping up.
	 */
	 
	if(ioctl(w->TVscreen.fbttv, VIDIOCSWIN, &w->TVscreen.vwin))
		perror("set clipping");
	return ncr;
}

static Boolean onscreen(w,wts)
TVscreenWidget w;
XWindowAttributes wts;
{
	int x,x2,y,y2;
	uint ww,wh;
	struct video_window vwin;
  
	vwin=w->TVscreen.vwin;
	wh=vwin.height; ww=vwin.width;

	if (!(wts.map_state & IsViewable))
		return False;

	x=wts.x-vwin.x; y=wts.y-vwin.y;
	x2=x+wts.width+2*wts.border_width-1;
	y2=y+wts.height+2*wts.border_width-1;
	if ((x2 < 0) || (x >= (int)ww) ||
		(y2 < 0) || (y >= (int)wh))
		return False;
	return True;
}

static void cap(w,W)
TVscreenWidget w;
int W;
{
	if( XtIsRealized((Widget)w)) 
	{
		int one=1;
		int zero=0;
		if(W && w->TVscreen.CAP_ON)
		{
			if(ioctl(w->TVscreen.fbttv, VIDIOCCAPTURE, &one))
				perror("capture on");
		}
		else
		{
			if(ioctl(w->TVscreen.fbttv, VIDIOCCAPTURE, &zero))
				perror("capture off");
		}
	}
}

static void setmaxsize(cw)
TVscreenWidget cw;
{
	uint max_h,max_w;

	switch (cw->TVscreen.NORM)
	{
		case PAL:
		case SECAM:
			max_w=PAL_WIDTH;
			max_h=PAL_HEIGHT;
			break;

		case NTSC:
			max_w=NTSC_WIDTH;
			max_h=NTSC_HEIGHT;
			break;
    
		default:
			max_w=DEFAULT_WIDTH;
			max_h=DEFAULT_HEIGHT;
			break;
	}

	cw->TVscreen.max_width=max_w;
	cw->TVscreen.max_height=max_h;
}
  
static void Initialize(treq, tnew, args, num_args)
Widget treq, tnew;
ArgList args;
Cardinal *num_args;
{
	Widget parent;
	char *bttvname;
	Screen *screen;
	Window rootw;
	int col;
	uint rwidth; 
	TVscreenWidget new = (TVscreenWidget) tnew;
	int major, minor, bank, ram;
	Display *disp;
	ushort i,r,rr;
	Colormap cmap;
	XColor *my;
	Visual *v;

	/*
	 *	FIXME: /dev/video00->xx
	 *	Move all the vbuf stuff to the Xserver.
	 */
	if(new->TVscreen.devname)
		bttvname=new->TVscreen.devname;
	else 
		bttvname="/dev/bttv";

	if ((new->TVscreen.fbttv=open(bttvname,O_RDWR)) < 0) 
	{
		fprintf(stderr,"Cannot open %s\n",bttvname);
		exit(1);
	}

	if(ioctl(new->TVscreen.fbttv, VIDIOCGWIN, &(new->TVscreen.vwin)))
		perror("get window");
	if(ioctl(new->TVscreen.fbttv, VIDIOCGFBUF, &(new->TVscreen.vwin)))
		perror("get frame buffer");
	if(ioctl(new->TVscreen.fbttv, VIDIOCGTUNER, &(new->TVscreen.vtuner)))
		perror("get tuner");
	new->TVscreen.vwin.width=DEFAULT_WIDTH;
	new->TVscreen.vwin.height=DEFAULT_HEIGHT;
	new->TVscreen.vwin.x=0;
	new->TVscreen.vwin.y=0;
	new->TVscreen.vbuf.width=XtScreen(tnew)->width;
	new->TVscreen.vbuf.height=XtScreen(tnew)->height;
	new->TVscreen.vwin.clipcount=0;
	new->TVscreen.vwin.width=32;
	new->TVscreen.vwin.height=32;

	rwidth=new->TVscreen.vbuf.width;
	disp=XtDisplay(tnew);
	if (XF86DGAQueryVersion(disp, &major, &minor)) 
		XF86DGAGetVideoLL(disp, DefaultScreen(disp), &(new->TVscreen.vbuf.base),
		&rwidth, &bank, &ram);
	else 
	{
		if (rwidth<960)
		{
			if (rwidth>640)
				rwidth=960;
		}
	}
	if (new->TVscreen.realw)
		rwidth=new->TVscreen.realw;
	new->TVscreen.dual=find_visual(tnew,new->TVscreen.visname,
		 &new->TVscreen.visual);

	v= new->TVscreen.visual;
	if (new->TVscreen.bpp)
		new->TVscreen.vbuf.depth=new->TVscreen.bpp;
	else
		new->TVscreen.vbuf.depth=tnew->core.depth;
	if (new->TVscreen.dual)
		new->TVscreen.vbuf.depth=32;
	new->TVscreen.vbuf.bytesperline=rwidth*new->TVscreen.vbuf.depth/8;
	
	if(ioctl(new->TVscreen.fbttv, VIDIOCSFBUF, &new->TVscreen.vbuf))
	{
		perror("set frame buffer");
		exit(1);
	}

	if (v->class == PseudoColor)
	{
		my=new->TVscreen.colors;
		for (i=0;i<256;i++)
		{
			my[i].pixel=i;
			my[i].flags = DoRed | DoGreen | DoBlue;
		}
		XQueryColors(disp, new->core.colormap, my, 256);

		cmap=XCreateColormap(disp, DefaultRootWindow(disp),v, AllocAll);
		new->core.colormap=cmap;
		for (i=0;i<225;i++)
		{ /* Set Colors 16 - 240 */
			r=i%25;
			my[i+16].green =(65535*(i/25)/8);  
			my[i+16].red =(65535*(r/5)/4);  
			rr=r%5;
			my[i+16].blue =(65535*rr/4);        
		}
		XStoreColors(disp,cmap,my,256);
	}

	new->core.width=DEFAULT_WIDTH;
	new->core.height=DEFAULT_HEIGHT; 


	parent=tnew;
	while (parent->core.parent)
	{
		parent=parent->core.parent;
	}

	XtAddEventHandler(tnew,SubstructureNotifyMask | VisibilityChangeMask
			| StructureNotifyMask | ExposureMask
			| EnterWindowMask |LeaveWindowMask,
			FALSE,pConfig,new);
	XtAddEventHandler(parent,SubstructureNotifyMask | VisibilityChangeMask
			| StructureNotifyMask | ExposureMask,
			FALSE,pConfig,new);
	rootw=DefaultRootWindow(XtDisplay(tnew));
	XSelectInput(XtDisplay(tnew),rootw, VisibilityChangeMask |
		SubstructureNotifyMask | ExposureMask |StructureNotifyMask);
  
  
	XtRegisterDrawable(XtDisplay(tnew),rootw,tnew);



	new->TVscreen.Change=FALSE;
	setmaxsize(new);
	new->core.background_pixel=BlackPixel(XtDisplay(new),
					DefaultScreen(XtDisplay(new)));

	timer=0;
	XSetErrorHandler((XErrorHandler) xerror);
}


static void SetGeo(cw)
TVscreenWidget cw;
{
	Widget parent;
	Position x,y;
	int dx,dy;
	Boolean CHECK;
	int oldx,oldy,oldw,oldh;

	oldx=cw->TVscreen.vwin.x;
	oldy=cw->TVscreen.vwin.y;
	oldw=cw->TVscreen.vwin.width;
	oldh=cw->TVscreen.vwin.height;

	CHECK=cw->TVscreen.Change;

	cw->TVscreen.vwin.flags=0;

	if (!XtIsRealized((Widget) cw))
		return;

	parent=(Widget) cw;
	while(XtParent(parent))
		parent=XtParent(parent);
  
	XtTranslateCoords((Widget)cw,0,0,&x,&y);
 
	if(oldh!=cw->core.height)
		CHECK=TRUE;
	if(oldw!=cw->core.width)
		CHECK=TRUE;

	dx=0;
	dy=0;
	if (cw->core.width > cw->TVscreen.max_width)
	{
		dx=(cw->core.width- oldw)/2;
		if(oldw != cw->TVscreen.max_width)
		{
			CHECK=TRUE;
			cw->TVscreen.vwin.width=cw->TVscreen.max_width & ~3;
		}
	}
	else
	{
		cw->TVscreen.vwin.width=cw->core.width & ~3;
	}

	if (cw->core.height > cw->TVscreen.max_height)
	{
		 dy=(cw->core.height- cw->TVscreen.vwin.height)/2;
		if(oldh != cw->TVscreen.max_height)
		{
			CHECK=TRUE;
			cw->TVscreen.vwin.height=cw->TVscreen.max_height & ~3;
		}
	}
	else
		cw->TVscreen.vwin.height=cw->core.height & ~3;

	if(oldx!=x+dx)
	{
		Move=True;
		CHECK=True;
	}
	if(oldy!=y+dy)
	{
		Move=True;
		CHECK=True;
	}
	
	/*
	 *	FIXME - need a "GETGEOM" type way to get the ideal geometry
	 */

	if ((cw->TVscreen.vbuf.depth==24)&&((x+dx)&3)) 
	{
		parent->core.x=parent->core.x-((x+dx) &3);
	}
	else if ((cw->TVscreen.vbuf.depth==16)&&((x+dx)&1))
	{
		parent->core.x=parent->core.x-((x+dx) &1);
	}
	if ((y+dy)&1) 
	{
		parent->core.y=parent->core.y-((y+dy) &1);
	}
  
	if(CHECK)
	{
		cw->TVscreen.vwin.x=x+dx;
		cw->TVscreen.vwin.y=y+dy;
		if (cw->TVscreen.CAP_ON)
		{
			if(ioctl(cw->TVscreen.fbttv, VIDIOCSWIN, &cw->TVscreen.vwin))
				perror("set clipped window");
		}
	}  
	if(cw->TVscreen.CAP_ON &&  CHECK)
	{
		getclip(cw); 
	}
	if(!cw->TVscreen.Change && cw->TVscreen.CAP_ON && cw->TVscreen.Fresh) 
	{
		refresh(cw);
		cw->TVscreen.Fresh=False;
	}
	cw->TVscreen.Change=FALSE;
}



static Boolean
SetValues(current, request, new, args, num_args)
Widget current, request, new;
ArgList args;
Cardinal *num_args;
{
	Widget parent;
	TVscreenWidget curcw = (TVscreenWidget) current;
	TVscreenWidget newcw = (TVscreenWidget) new;
	Boolean do_redisplay = False;
	struct video_picture vp;
	int pictmod=0;

	if (newcw->TVscreen.CAP_ON)
	{
		if (!curcw->TVscreen.CAP_ON) 
		{
			do_redisplay =True;
			newcw->TVscreen.Change=TRUE;
			SetGeo(newcw);
		}
	}
	else 
	{
		int v=newcw->TVscreen.INPUT;
		if (curcw->TVscreen.CAP_ON) 
		{
			cap(newcw,0);
			system("xrefresh");
		}
		if(ioctl(newcw->TVscreen.fbttv, VIDIOCSCHAN, &v))
			perror("change tuner");
	}

	if (newcw->TVscreen.frequency!=(ushort) newcw->TVscreen.FREQ)
	{
		newcw->TVscreen.frequency=newcw->TVscreen.FREQ;
		if(ioctl(newcw->TVscreen.fbttv, VIDIOCSFREQ, &newcw->TVscreen.frequency))
			perror("change frequency");
	}
  
  	/*
  	 *	Need to create a set format ioctl ??
  	 */
  	 
	if (newcw->TVscreen.vtuner.mode != newcw->TVscreen.NORM)
	{
		newcw->TVscreen.vtuner.mode=newcw->TVscreen.NORM;
		newcw->TVscreen.Change=TRUE;
		setmaxsize(newcw);
		SetGeo(newcw);
		if (newcw->TVscreen.NORM==NTSC)
			newcw->TVscreen.Fresh=TRUE;
		SetGeo(newcw);
		if(ioctl(newcw->TVscreen.fbttv, VIDIOCSTUNER, &newcw->TVscreen.vtuner))
			perror("change format");
	}

	if (curcw->TVscreen.INPUT !=newcw->TVscreen.INPUT)
	{
		int v=newcw->TVscreen.INPUT;
		if(ioctl(newcw->TVscreen.fbttv, VIDIOCSCHAN, &v))
			perror("change input");
	}

	if(ioctl(newcw->TVscreen.fbttv, VIDIOCGPICT, &vp)==-1)
		perror("get picture params");
  
	if (curcw->TVscreen.Bright !=newcw->TVscreen.Bright)
	{
		vp.brightness=(newcw->TVscreen.Bright+128)<<8;
		pictmod=1;
	}

	if (curcw->TVscreen.Color !=newcw->TVscreen.Color)
	{
		vp.colour=newcw->TVscreen.Color<<7;
		pictmod=1;
	}

	if (curcw->TVscreen.Hue !=newcw->TVscreen.Hue)
	{
		vp.hue=(newcw->TVscreen.Hue+128)<<8;
		pictmod=1;
	}


	if (curcw->TVscreen.Contrast !=newcw->TVscreen.Contrast)
	{
		vp.contrast=newcw->TVscreen.Contrast<<7;
		pictmod=1;
	}
  
	if(pictmod)
	{
		if(ioctl(newcw->TVscreen.fbttv, VIDIOCSPICT, &vp))
			perror("set picture params");
	}  

	return do_redisplay;
}







static void pConfig(wi,client_data,event)
     Widget wi;
     XtPointer client_data;
     XEvent *event;
{
  int fbttv;
  TVscreenWidget w=(TVscreenWidget) client_data;
  static Boolean UNMAPPED=False;
  XWindowAttributes wts1,wts2;
  Window checkw1,checkw2,wwin;
  Widget parent;

	if (!XtIsRealized((Widget)w)||(wwin=XtWindow((Widget)w)) == None)
  		return;

	parent=(Widget)w;
	while (parent->core.parent)
	{
		parent=parent->core.parent;
	}
  
	fbttv=w->TVscreen.fbttv;
	w->TVscreen.Fresh=True;
	cap(w,0);

	checkw1=event->xmap.window;
	 checkw2=event->xany.window;
	if (w->TVscreen.CAP_ON)
	{ 
		if ( wwin==checkw1 || wwin==checkw2 )
			w->TVscreen.Fresh=False;
		else 
		{
			if (checkw1)
			{ 
				XGetWindowAttributes(XtDisplay(w), checkw1, &wts1);
				if(!onscreen(w,wts1))
					w->TVscreen.Fresh=False; 
			}
			if (checkw2)
			{
				XGetWindowAttributes(XtDisplay(w), checkw2, &wts2);
				if(!onscreen(w,wts2))
					w->TVscreen.Fresh=False; 
			}
		}
	}
  
	switch (event->type) 
	{
		case VisibilityNotify:
			switch (event->xvisibility.state) 
			{
				case VisibilityUnobscured:
					if (w->TVscreen.CAP_ON)
					{
						if (checkw2 != DefaultRootWindow(XtDisplay(w))) 
						{
							w->TVscreen.Change=TRUE;
							SetGeo(w);
						}
					}
					break;
				case VisibilityFullyObscured:
					if (checkw2 != DefaultRootWindow(XtDisplay(w))) 
						cap(w,0);
					break;
				case VisibilityPartiallyObscured:
					if (w->TVscreen.CAP_ON)
					{
						if (checkw2 != DefaultRootWindow(XtDisplay(w))) 
						{
							w->TVscreen.Change=TRUE;
							SetGeo(w);
						}
					}
					break;
			}
			break;

		case MapNotify:
			if ((DefaultRootWindow(XtDisplay(w)) != checkw2) && UNMAPPED)
			{
				UNMAPPED=False;
				w->TVscreen.CAP_ON=True;
				cap(w,1);
				if ( w->TVscreen.Fresh )
				{
					refresh(w);
				}
			}
			if ((lastrwin != checkw1) && (w->TVscreen.CAP_ON) )
			{
				if ( w->TVscreen.Fresh )
				{
					refresh(w);
				}
			}
			break;


		case ConfigureNotify:
			cap(w,0);
			SetGeo(w);
			break;

		case DestroyNotify:
			if (wwin  == checkw2)
			{
				/* Erm shouldnt we cap ,0 always ?? AC */
				cap(w,0);
				XFreeColormap(XtDisplay(w),w->core.colormap);
#ifdef DEBUG
				printf("Destroy\n");
#endif
			}
			break;
			
		case UnmapNotify:
			if (DefaultRootWindow(XtDisplay(w)) != checkw2)
			{
				cap(w,0);
				if (w->TVscreen.CAP_ON)
					UNMAPPED=True;
				if (w->TVscreen.CAP_ON) 
				{
					if ( w->TVscreen.Fresh )
					{
						refresh(w);
#ifdef DEBUG
						printf("UnmapNotify refresh\n");
#endif
					}
				}
				w->TVscreen.CAP_ON=False;
#ifdef DEBUG
				printf("Unmap\n");
#endif
			}
			else if (wwin  != event->xmap.window)
			{
				if(w->TVscreen.CAP_ON) 
				{
					getclip(w); 
				}
			}
			break;
  
		case EnterNotify:
			/*
			 *	Handle private colour
			 */
			if (wwin  == checkw2)
			{
				if (w->TVscreen.CAP_ON)
				{ 
					XInstallColormap(XtDisplay(w),w->core.colormap);
				}
			}
			break;

		case LeaveNotify:
			if (wwin  == checkw2)
			{
				if (w->TVscreen.CAP_ON)
				{ 
					XUninstallColormap(XtDisplay(w),w->core.colormap);
				}
			}
			break;
    
	}
  
	if(w->TVscreen.CAP_ON) 
	{
		cap(w,1); 
	}
}

static XtGeometryResult QueryGeometry(w, proposed, answer)
     Widget w;
     XtWidgetGeometry *proposed, *answer;
{
	int reqm;
	TVscreenWidget cw = (TVscreenWidget) w;
  
	/* set fields we care about */
	reqm= CWWidth | CWHeight;
	answer->request_mode = reqm;
  
	/* initial width and height */
	answer->width = (cw->TVscreen.vwin.width > 
		cw->TVscreen.max_width) ? cw->TVscreen.max_width : 
					  cw->TVscreen.vwin.width;
	answer->height = (cw->TVscreen.vwin.height > 
	    cw->TVscreen.max_height) ? cw->TVscreen.max_height : 
				       cw->TVscreen.vwin.height;
  
	if (cw->TVscreen.vbuf.depth==24) 
	{
		answer->width = answer->width & ~3; 
		answer->height = answer->height & ~3; 
	}
	else if (cw->TVscreen.vbuf.depth==16)
	{
		answer->width = answer->width & ~1; 
		answer->height = answer->height & ~1; 
	}

	if (  ((proposed->request_mode & (reqm)) == (reqm)) &&
		proposed->width == answer->width &&
		proposed->height == answer->height)
		return XtGeometryYes;
	else if (answer->width == cw->core.width &&
		answer->height == cw->core.height)
		return XtGeometryNo;
	else
		return XtGeometryAlmost;
}

static void destroy(w)
TVscreenWidget w;
{
	w->TVscreen.CAP_ON=False;
	cap(w,0);
	refresh(w);
#ifdef DEBUG
	printf("destroy refresh\n");
#endif
}
